/*

  yyycpu - a virtual machine
  Copyright (C) 2000  Erik Thiele

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "membunch.hpp"

#include "fatal.hpp"

#include <stdio.h>

static const string numberu32(const uint32 x)
{
  char s[100];
  sprintf(s,"%ud",x);
  return s;
};

membunch::membunch (uint8 *const memory_, const uint32 memsize_)
  : memory(memory_), memsize(memsize_)
{
};

uint8 membunch::get8(const uint32 addr) const
{
  if (addr<memsize) {
    return memory[addr];
  }else{
    warning("membunch::get8("+numberu32(addr)+"): address out of range. returning 0");
    return 0;
  };
};

void membunch::set8(const uint32 addr, const uint8 value)
{
  if (addr<memsize) {
    memory[addr]=value;
  }else{
    warning("membunch::get8("+numberu32(addr)+"): address out of range. doing nothing.");
  };
};

uint16 membunch::get16(const uint32 addr) const
{
  if (addr < memsize-1) {
    return (((uint16)memory[addr])<<8) | (uint16)memory[addr+1];
  }else{
    warning("membunch::get16("+numberu32(addr)+"): address out of range. returning 0");
    return 0;
  };
};

void membunch::set16(const uint32 addr, const uint16 value)
{
  if (addr < memsize-1) {
    memory[addr] = value>>8;
    memory[addr+1] = value&255;
  }else{
    warning("membunch::get16("+numberu32(addr)+"): address out of range. doing nothing.");
  };
};

uint32 membunch::get32(const uint32 addr) const
{
  if (addr < memsize-3) {
    return (((uint32)memory[addr])<<24) | (((uint32)memory[addr+1])<<16) | 
      (((uint32)memory[addr+2])<<8) | (uint32)memory[addr+3];
  }else{
    warning("membunch::get32("+numberu32(addr)+"): address out of range. returning 0");
    return 0;
  };
};

void membunch::set32(const uint32 addr, const uint32 value)
{
  if (addr < memsize-3) {
    memory[addr]=value>>24;
    memory[addr+1]=(value>>16)&255;
    memory[addr+2]=(value>>8)&255;
    memory[addr+3]=value&255;
  }else{
    warning("membunch::get32("+numberu32(addr)+"): address out of range. doing nothing");
  };
};
