%{
#define YY_NO_UNPUT
#include "../common/binarydefs.hpp"
#include "../common/hexconvert.hpp"
#include "global.hpp"
#include "parser.h"
#include <errno.h>

constlist::elem::value_t parsenumber(char *s, const int base)
{
  errno=0;
  constlist::elem::value_t x=strtoll (s,0,base);
  if (errno==ERANGE) yyerror("number parse overflow");
  return x;
};

%}

%x COMMENT
%x TERMINATOR

%%
NOP		{return COMMAND_NOP;}
PUSHCW		{return COMMAND_PUSHCW;}
PUSHCD		{return COMMAND_PUSHCD;}
JMPCW		{return COMMAND_JMPCW;}
JMPCD		{return COMMAND_JMPCD;}
CALLCW		{return COMMAND_CALLCW;}
CALLCD		{return COMMAND_CALLCD;}
RET		{return COMMAND_RET;}
IRET		{return COMMAND_IRET;}
HALT		{return COMMAND_HALT;}
CONSW		{return COMMAND_CONSW;}
CONSD		{return COMMAND_CONSD;}
PUSHR		{return COMMAND_PUSHR;}
POPR		{return COMMAND_POPR;}
JMPR		{return COMMAND_JMPR;}
CALLR		{return COMMAND_CALLR;}
IN		{return COMMAND_IN;}
OUT		{return COMMAND_OUT;}
INC		{return COMMAND_INC;}
DEC		{return COMMAND_DEC;}
ADD		{return COMMAND_ADD;}
SUB		{return COMMAND_SUB;}
NEG		{return COMMAND_NEG;}
SMUL		{return COMMAND_SMUL;}
SDIV		{return COMMAND_SDIV;}
SHL		{return COMMAND_SHL;}
SHR		{return COMMAND_SHR;}
SAR		{return COMMAND_SAR;}
AND		{return COMMAND_AND;}
OR		{return COMMAND_OR;}
XOR		{return COMMAND_XOR;}
NOT		{return COMMAND_NOT;}
MOV		{return COMMAND_MOV;}
READCW		{return COMMAND_READCW;}
READCD		{return COMMAND_READCD;}
WRITCW		{return COMMAND_WRITCW;}
WRITCD		{return COMMAND_WRITCD;}
XSGCW		{return COMMAND_XSGCW;}
XSGCD		{return COMMAND_XSGCD;}
XSGECW		{return COMMAND_XSGECW;}
XSGECD		{return COMMAND_XSGECD;}
XECW		{return COMMAND_XECW;}
XECD		{return COMMAND_XECD;}
XNECW		{return COMMAND_XNECW;}
XNECD		{return COMMAND_XNECD;}
SHLC		{return COMMAND_SHLC;}
SHRC		{return COMMAND_SHRC;}
SARC		{return COMMAND_SARC;}
CALLCT		{return COMMAND_CALLCT;}
JMPCT		{return COMMAND_JMPCT;}
PUSHCT		{return COMMAND_PUSHCT;}
READR		{return COMMAND_READR;}
WRITR		{return COMMAND_WRITR;}
CONSB		{return COMMAND_CONSB;}

db		{return ASM_db;}
dw		{return ASM_dw;}
dd		{return ASM_dd;}

R0		{yylval.regist=0;return REGIST;}
R1		{yylval.regist=1;return REGIST;}
R2		{yylval.regist=2;return REGIST;}
R3		{yylval.regist=3;return REGIST;}
R4		{yylval.regist=4;return REGIST;}
R5		{yylval.regist=5;return REGIST;}
R6		{yylval.regist=6;return REGIST;}
R7		{yylval.regist=7;return REGIST;}
R8		{yylval.regist=8;return REGIST;}
R9		{yylval.regist=9;return REGIST;}
R10		{yylval.regist=10;return REGIST;}
R11		{yylval.regist=11;return REGIST;}
R12		{yylval.regist=12;return REGIST;}
R13		{yylval.regist=13;return REGIST;}
R14		{yylval.regist=14;return REGIST;}
R15		{yylval.regist=15;return REGIST;}

[01]+b		{yylval.constnormal=parsenumber(yytext,2); return CONSTNORMAL;}
[0-7]+o		{yylval.constnormal=parsenumber(yytext,8); return CONSTNORMAL;}
[0-9]+n		{yylval.constnormal=parsenumber(yytext,10); return CONSTNORMAL;}
[A-F0-9]+h	{yylval.constnormal=parsenumber(yytext,16); return CONSTNORMAL;}

\"[ -\!\#-~]*\"	{char t=yytext[yyleng-1];yytext[yyleng-1]=0;
		yylval.strgindex=strgl->add_interpreted(yytext+1);yytext[yyleng-1]=t; return STRGINDEX;}

[a-z_][a-z0-9_]*:	{char t=yytext[yyleng-1];yytext[yyleng-1]=0;
			ll->defineoradd(sl->findoradd(yytext),currentaddress);yytext[yyleng-1]=t;}

[a-z_][a-z0-9_]*	{yylval.labelindex=ll->findoraddundefined(sl->findoradd(yytext)); return LABELINDEX;}

<INITIAL><<EOF>>	{BEGIN(TERMINATOR); return ENDOFLINE;}
\n		{return ENDOFLINE;}
,		{return COMMA;}
:		{return COLON;}
\+		{return PLUS;}
\-		{return MINUS;}
\*		{return MULTIPLY;}
\/		{return DIVIDE;}
\(		{return BRACKETOPEN;}
\)		{return BRACKETCLOSE;}
[\t ]		;

;		BEGIN(COMMENT);
<COMMENT>.+	;
<COMMENT>\n	{BEGIN(INITIAL); return ENDOFLINE;}
<COMMENT><<EOF>>	{BEGIN(TERMINATOR); return ENDOFLINE;}

[a-zA-Z0-9_]+	{yyerror("wrong command");}
.		{yyerror("wrong character");}
%%

void yyerror(char *s)
{
  fprintf(stderr,"ERROR: %s: %s\n",s,yytext);
  exit(1);
};
