/*

  yyycpu - a virtual machine
  Copyright (C) 2000  Erik Thiele

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "machcore.h"

#include <stdlib.h>

static u_int32_t limit16;
static u_int32_t limit32;

#define B0000 0
#define B0001 1
#define B0010 2
#define B0011 3
#define B0100 4
#define B0101 5
#define B0110 6
#define B0111 7
#define B1000 8
#define B1001 9
#define B1010 10
#define B1011 11
#define B1100 12
#define B1101 13
#define B1110 14
#define B1111 15

#define memsize machcore_memsize
#define memory machcore_memory
#define PC machcore_PC
#define regs machcore_regs
#define interrupt_enable machcore_interrupt_enable

static u_int8_t nib0(const u_int16_t value) { return value&15; };
static u_int8_t nib1(const u_int16_t value) { return (value>>4)&15; };
static u_int8_t nib2(const u_int16_t value) { return (value>>8)&15; };
static u_int8_t nib3(const u_int16_t value) { return value>>12; };

static u_int8_t byt0(const u_int16_t value) { return value&255; };
/*static u_int8_t byt1(const u_int16_t value) { return value>>8; };*/

static u_int16_t get16bit (const u_int32_t addr)
{
  if (addr <= limit16) {
    return (((u_int16_t)memory[addr])<<8) | (u_int16_t)memory[addr+1];
  }else{
    machcore_illegalmemoryaccess();
    return 0;
  };
};

/*static void set16bit (const u_int32_t addr, const u_int8_t val)
{
 if (addr <= limit16) {
   memory[addr] = val>>8;
   memory[addr+1] = val&255;
 } else machcore_illegalmemoryaccess();
};*/

static u_int32_t get32bit(const u_int32_t addr)
{
  if (addr <= limit32) {
    return (((u_int32_t)memory[addr])<<24) | (((u_int32_t)memory[addr+1])<<16) | 
      (((u_int32_t)memory[addr+2])<<8) | (u_int32_t)memory[addr+3];
  }else{
    machcore_illegalmemoryaccess();
    return 0;
  };
};

/*static void set32bit(const u_int32_t addr, const u_int8_t val)
{
  if (addr <= limit16) {
    memory[addr]=val>>24;
    memory[addr+1]=(val>>16)&255;
    memory[addr+2]=(val>>8)&255;
    memory[addr+3]=val&255;
  } else machcore_illegalmemoryaccess();
};*/

static int32_t getdist16(const u_int32_t addr)
{
  return ((int32_t)(int16_t)(get16bit(addr))) << 1;
};

static int32_t getdist32(const u_int32_t addr)
{
  return get32bit(addr) << 1;
};

static int32_t getdist12(u_int32_t addr)
{
  int32_t a = get16bit(addr) & 0xFFF;
  if (a & 0x800) a |= 0xFFFFF000;
  return a<<1;
};

void machcore_initialize(const u_int32_t memsize_)
{
  memsize=memsize_;
  limit16=memsize-2;
  limit32=memsize-4;
  memory=(u_int8_t*)malloc(memsize);
  machcore_reset();
};

void machcore_reset()
{
  int i;

  PC=0; // FIXME prg must start at 256
  for (i=0;i<16;++i) regs[i]=0;
  interrupt_enable=0;
};

void machcore_deinitialize()
{
  free(memory);
  memory=0;
};

void machcore_execute(int count)
{
  for (;count;--count) { // outer for loop

  u_int16_t ins=get16bit(PC);
  switch (nib3(ins)) {
  case B0000:
    switch (nib2(ins)) {
    case B0000:
      switch (nib1(ins)) {
      case B0000:
	switch (nib0(ins)) {
	case B0000: PC+=2; break;
	default: machcore_illegalinstruction(); break;
	};
	break;
      case B0001: regs[nib0(ins)] = get16bit(PC+2); PC+=4; break;
      case B0010: regs[nib0(ins)] = get32bit(PC+2); PC+=6; break;
      default: machcore_illegalinstruction(); break;
      };
      break;
    case B0001: regs[nib0(ins)] = machcore_in (regs[nib1(ins)]); PC+=2; break;
    case B0010: machcore_out (regs[nib1(ins)], regs[nib0(ins)]); PC+=2; break;
    case B0011: regs[nib1(ins)] = regs[nib0(ins)] + 1; PC+=2; break;
    case B0100: regs[nib1(ins)] = regs[nib0(ins)] - 1; PC+=2; break;
    case B0101: regs[nib1(ins)] += regs[nib0(ins)]; PC+=2; break;
    case B0110: regs[nib1(ins)] -= regs[nib0(ins)]; PC+=2; break;
    default: machcore_illegalinstruction(); break;
    };
    break;
  case B0001:
    switch (nib2(ins)) {
    case B0001: regs[nib1(ins)]=regs[nib0(ins)]; PC+=2; break;
    case B0010: regs[nib1(ins)]=get32bit(regs[nib0(ins)]+(u_int32_t)get16bit(PC+2)); PC+=4; break;
    case B0110: PC += (int32_t)regs[nib1(ins)] > (int32_t)regs[nib0(ins)] ? getdist16(PC+2) : 4; break;
    case B0111: PC += (int32_t)regs[nib1(ins)] > (int32_t)regs[nib0(ins)] ? getdist32(PC+2) : 6; break;
    case B1000: PC += (int32_t)regs[nib1(ins)] >= (int32_t)regs[nib0(ins)] ? getdist16(PC+2) : 4; break;
    case B1010: PC += (int32_t)regs[nib1(ins)] == (int32_t)regs[nib0(ins)] ? getdist16(PC+2) : 4; break;
    case B1100: PC += (int32_t)regs[nib1(ins)] != (int32_t)regs[nib0(ins)] ? getdist16(PC+2) : 4; break;
    default: machcore_illegalinstruction(); break;
    };
    break;
  case B0010:
    switch (nib2(ins) & B1110) {
    case B0010: regs[nib0(ins)] >>= (ins >> 4) & 31; PC+=2; break;
    case B0100: (int32_t)regs[nib0(ins)] >>= (ins >> 4) & 31; PC+=2; break;
    default: machcore_illegalinstruction(); break;
    };
    break;
  case B1011: PC += getdist12(PC); break;
  case B1111: regs[nib2(ins)]=byt0(ins); PC+=2; break;
  default: machcore_illegalinstruction(); break;
  };

  }; // outer for loop
};

u_int32_t memsize;
u_int8_t *memory = 0;
u_int32_t PC;
u_int32_t regs[16];
int interrupt_enable;
